package pico.orange.compiler;

import java.util.HashMap;
import java.util.Map;

import pico.orange.compiler.Token.TokenValue;

public class Lexer {

	private static Map<String, Token.TokenValue> KEYWORDS = new HashMap<String, Token.TokenValue>() {
		{
			this.put("org", TokenValue.ORG);
			this.put("dn", TokenValue.DN);
			this.put("ink", TokenValue.INK);
			this.put("outn", TokenValue.OUTN);
			this.put("abyz", TokenValue.ABYZ);
			this.put("ay", TokenValue.AY);
			this.put("st", TokenValue.ST);
			this.put("ld", TokenValue.LD);
			this.put("add", TokenValue.ADD);
			this.put("sub", TokenValue.SUB);
			this.put("ldi", TokenValue.LDI);
			this.put("addi", TokenValue.ADDI);
			this.put("ldyi", TokenValue.LDYI);
			this.put("addyi", TokenValue.ADDYI);
			this.put("cpi", TokenValue.CPI);
			this.put("cpyi", TokenValue.CPYI);
			this.put("scall", TokenValue.SCALL);
			this.put("jmpf", TokenValue.JMPF);
			this.put("call", TokenValue.CALL);
			this.put("ret", TokenValue.RET);
			this.put("pusha", TokenValue.PUSHA);
			this.put("popa", TokenValue.POPA);
			this.put("pushb", TokenValue.PUSHB);
			this.put("popb", TokenValue.POPB);
			this.put("pushy", TokenValue.PUSHY);
			this.put("popy", TokenValue.POPY);
			this.put("pushz", TokenValue.PUSHZ);
			this.put("popz", TokenValue.POPZ);
			this.put("ioctrl", TokenValue.IOCTRL);
			this.put("out", TokenValue.OUT);
			this.put("in", TokenValue.IN);
		}
	};

	private String src;
	private int pos;
	private int srcSize;
	private String stringValue;
	private int intValue;
	private int line;

	/**
	 * 字句解析を構築します。
	 *
	 * @param src
	 *            ソース
	 */
	public Lexer(String src) {
		this.src = src;
		this.srcSize = src.length();
		this.line = 1;
	}

	/**
	 * 次のトークンを取得します。
	 *
	 * @return トークン
	 */
	public Token nextToken() {
		TokenValue tokenValue = this.nextTokenValue();
		return new Token(tokenValue, this.stringValue, this.intValue, this.line);
	}

	/**
	 * トークンを先読みします。nextToken()で取得するトークンに影響を与えません。
	 *
	 * @param n
	 *            先読み個数
	 * @return トークン
	 */
	public Token lookToken(int n) {
		TokenValue tokenValue = null;
		int pos = this.pos;
		int line = this.line;
		for (int i = 0; i < n; i++) {
			tokenValue = this.nextTokenValue();
		}
		Token token = new Token(tokenValue, this.stringValue, this.intValue, this.line);
		this.pos = pos;
		this.line = line;
		return token;
	}

	/**
	 * 次のトークンを取得します。
	 * 
	 * @return トークン値
	 */
	private TokenValue nextTokenValue() {
		while (true) {
			int c;
			do {
				c = this.getCh();
			} while (this.isWhitespace(c));
			if (this.isLetter(c)) {
				StringBuilder buf = new StringBuilder();
				while (this.isLetter(c) || this.isDigit(c)) {
					buf.append((char)c);
					c = this.getCh();
				}
				if (c != -1) {
					this.ungetCh();
				}
				this.stringValue = buf.toString();

				// キーワードテーブルから検索
				TokenValue tokenValue = Lexer.KEYWORDS.get(this.stringValue);
				if (tokenValue == null) {
					tokenValue = TokenValue.IDENT;
				}
				return tokenValue;
			} else if (this.isDigit(c)) {
				StringBuilder buf = new StringBuilder();
				if (c == '0') {
					c = this.getCh();
					if (c == 'x' || c == 'X') { // 16進数
						c = this.getCh();
						while (this.isXdigit(c)) {
							buf.append((char)c);
							c = this.getCh();
						}
						this.ungetCh();
						this.intValue = Integer.parseInt(buf.toString(), 16);
						return TokenValue.LITERAL;
					} else {
						this.ungetCh();
						c = '0';
					}
				}
				while (this.isDigit(c)) { // 10進数
					buf.append((char)c);
					c = this.getCh();
				}
				ungetCh();
				this.intValue = Integer.parseInt(buf.toString(), 10);
				return TokenValue.LITERAL;
			}
			switch (c) {
			case ':':
				return TokenValue.COLON;
			case ';':
				do { // 行末までスキップ
					c = this.getCh();
				} while (c != '\n' && c != -1);
				break;
			case -1:
				return TokenValue.EOF;
			default:
				return TokenValue.ERROR;
			}
		}
	}

	private int getCh() {
		int c = (this.pos < this.srcSize) ? this.src.charAt(this.pos++) : -1;
		if (c == '\n') {
			this.line++;
		}
		return c;
	}

	private void ungetCh() {
		this.pos--;
		if (this.src.charAt(this.pos) == '\n') {
			this.line--;
		}
	}

	private boolean isWhitespace(int c) {
		return c == ' ' || c == '\t' || c == '\r' || c == '\n';
	}

	private boolean isLetter(int c) {
		return ('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z') || c == '_';
	}

	private boolean isDigit(int c) {
		return '0' <= c && c <= '9';
	}

	private boolean isXdigit(int c) {
		return ('0' <= c && c <= '9') || ('a' <= c && c <= 'f') || ('A' <= c && c <= 'F');
	}

}
