package pico.orange.ide.editor;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.JViewport;
import javax.swing.ScrollPaneConstants;
import javax.swing.border.Border;
import javax.swing.border.LineBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.text.Document;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledEditorKit;
import javax.swing.text.TabSet;
import javax.swing.text.TabStop;
import javax.swing.text.ViewFactory;
import javax.swing.undo.UndoManager;

import pico.orange.ide.Config;

public class Editor extends JPanel implements ChangeListener {

	private JTextPane textPane;
	private Header headPane;
	private JViewport editView;
	private JScrollPane editScroll;

	UndoManager undoManager = new UndoManager();

	public void undo() {
		if (this.undoManager.canUndo()) {
			this.undoManager.undo();
		}
	}

	public void redo() {
		if (this.undoManager.canRedo()) {
			this.undoManager.redo();
		}
	}

	public Editor() {
		super(new BorderLayout());

		this.textPane = new JTextPane();

		this.add(this.setDefaultTextPane(), BorderLayout.CENTER);
		this.setTabs(this.textPane, Config.TAB_SIZE);

		LineNumberView lineNumberView = new LineNumberView(this.textPane);
		this.editScroll.setRowHeaderView(lineNumberView);

		this.headPane = new Header();
		this.setDefaultHeadPane();

		this.setHeadPane();
		this.textPane.setUI(new LineHighlighter(this.textPane));

		// UndoManagerの登録
		Document doc = this.textPane.getDocument();
		doc.addUndoableEditListener(this.undoManager);

		// UNDO/REDOキーの登録
		this.textPane.addKeyListener(new KeyListener() {
			@Override
			public void keyTyped(KeyEvent e) {
			}

			@Override
			public void keyPressed(KeyEvent e) {
				if (e.isControlDown()) { // CTRLキー
					switch (e.getKeyCode()) {
					case KeyEvent.VK_Z:
						Editor.this.undo();
						e.consume();
						break;
					case KeyEvent.VK_Y:
						Editor.this.redo();
						e.consume();
						break;
					}
				}
			}

			@Override
			public void keyReleased(KeyEvent e) {
			}
		});
	}

	private JComponent setDefaultTextPane() {
		StyledEditorKit orgKit = new StyledEditorKit() {

			@Override
			public ViewFactory getViewFactory() {
				return new CustomViewFactory();
			}
		};
		this.textPane.setEditorKit(orgKit);
		this.textPane.setFont(Config.EDITOR_FONT);
		this.textPane.setBackground(Config.EDITOR_BG_COLOR); // エディターの背景色
		this.textPane.setCaretColor(Color.BLACK);

		this.textPane.setStyledDocument(new SyntaxDocument());

		this.editScroll = new JScrollPane(this.textPane, ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
				ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
		this.editScroll.setBackground(new Color(240, 240, 240));
		this.editScroll.setBorder(new LineBorder(Color.LIGHT_GRAY)); // エディターの外枠の色

		this.editView = this.editScroll.getViewport();
		this.editView.addChangeListener(this);

		return this.editScroll;
	}

	private void setDefaultHeadPane() {
		// headPane.setBorder(new LineBorder(Color.WHITE));
		Border border = BorderFactory.createMatteBorder(0, 0, 1, 0, Color.LIGHT_GRAY);
		this.headPane.setBorder(border);

		this.editScroll.setColumnHeaderView(this.headPane);
	}

	public JTextPane getTextPane() {
		return this.textPane;
	}

	public String getText() {
		return this.textPane.getText();
	}

	public void setText(String text) {
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				Editor.this.textPane.setText(text);
			}
		});
	}

	public void setTabs(JTextPane textpane, int tabSize) {
		FontMetrics fm = textpane.getFontMetrics(textpane.getFont());
		int charWidth = fm.charWidth('m');
		int tabLength = charWidth * tabSize;
		TabStop[] tabs = new TabStop[100];
		for (int j = 0; j < tabs.length; j++) {
			tabs[j] = new TabStop((j + 1) * tabLength);
		}
		TabSet tabSet = new TabSet(tabs);
		SimpleAttributeSet attrs = new SimpleAttributeSet();
		StyleConstants.setTabSet(attrs, tabSet);
		StyleConstants.setForeground(attrs, Color.BLACK);
		int l = textpane.getDocument().getLength();
		textpane.getStyledDocument().setParagraphAttributes(0, l, attrs, false);
	}

	public void setHeadPane() {
		this.headPane.setEditorFontInfo(this.textPane.getFont().getSize() / 2, this.textPane.getMargin());
		this.headPane.setPreferredSize(new Dimension(this.textPane.getWidth(), 12));
		this.headPane.repaint();
	}

	@Override
	public void stateChanged(ChangeEvent event) {
		int edit = this.editView.getViewPosition().x;
		this.headPane.setPreferredSize(new Dimension(this.editScroll.getWidth() + edit, this.headPane
				.getPreferredSize().height));
	}

}