package pico.orange.ide;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.ImageProducer;
import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;
import javax.swing.UIManager;

import pico.orange.compiler.Assembler;
import pico.orange.ide.console.Console;
import pico.orange.ide.editor.Editor;

/**
 *
 * @author pico
 *
 */
public class Frame extends JFrame implements DropTargetListener {

	private static final Color F_COLOR = new Color(240, 240, 240);
	private static final Color B_COLOR = new Color(240, 240, 240);

	private JTabbedPane editorTab;
	private File lastSelectedFile = new File(".");
	private File openFile;
	private Editor editor;
	private Console console;

	/**
	 * フレームを構築します。
	 */
	public Frame() {
		super();

		// 外観変更
		System.setProperty("swing.noxp", "true"); // JTabbedPane.setBackgroundAtを有効にする（Windows）
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
		}

		ImageIcon icon = this.getImageIcon("images/ORANGE4.png");
		this.setIconImage(icon.getImage());

		// タイトル
		this.setTitle("ORANGE-4 IDE");

		// メニューバー
		this.setJMenuBar(this.createMenuBar());

		// ツールバー
		JToolBar toolBar = this.createToolBar();
		toolBar.setBackground(Frame.B_COLOR);

		// 画面分割
		this.setSize(1024, 768);
		final JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
		splitPane.setBackground(Frame.B_COLOR);
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				splitPane.setDividerLocation(0.65);
			}
		});

		this.editorTab = new JTabbedPane(SwingConstants.TOP);
		splitPane.setLeftComponent(this.editorTab);
		JTabbedPane consoleTab = new JTabbedPane(SwingConstants.TOP);
		splitPane.setRightComponent(consoleTab);

		// 最小サイズ設定
		Dimension minimumSize = new Dimension(0, 0);
		consoleTab.setMinimumSize(minimumSize);

		// エディター
		this.editor = new Editor();
		this.editor.setFont(Config.EDITOR_FONT);
		this.editorTab.addTab("新規", this.editor);
		this.editorTab.setForegroundAt(0, Color.BLACK);
		this.editorTab.setBackgroundAt(0, Frame.F_COLOR);

		// コンソール
		this.console = new Console();
		consoleTab.setFont(Config.EDITOR_FONT);
		consoleTab.addTab("コンソール", this.console);
		consoleTab.setForegroundAt(0, Color.BLACK);
		consoleTab.setBackgroundAt(0, Frame.F_COLOR);

		// ステータスバー
		JLabel label = new JLabel(" ");

		// 配置
		Container contentPane = this.getContentPane();
		contentPane.setLayout(new BorderLayout());
		contentPane.add(toolBar, BorderLayout.NORTH);
		contentPane.add(splitPane, BorderLayout.CENTER);
		contentPane.add(label, BorderLayout.SOUTH);
		contentPane.setBackground(Frame.B_COLOR);

		// クローズボタン
		this.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				System.exit(0);
			}
		});
		
	    // ドラッグ＆ドロップ関連付け
//	    new DropTarget(this, this);
	    new DropTarget(this.editor.getTextPane(), this);
	    
		// 表示
		this.setLocationRelativeTo(null);
		this.setVisible(true);
	}

	/**
	 *
	 * @return
	 */
	private JMenuBar createMenuBar() {
		JMenuBar menuBar = new JMenuBar();

		// ファイル(F)
		JMenu menuFile = new JMenu("ファイル(F)");
		menuFile.setMnemonic('F');

		JMenuItem itemNew = new JMenuItem("新規(N)", 'N');
		itemNew.setAccelerator(KeyStroke.getKeyStroke('N', InputEvent.CTRL_MASK));
		itemNew.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.newFile();
			}
		});
		itemNew.setIcon(this.getImageIcon("images/new.png"));
		menuFile.add(itemNew);

		JMenuItem itemOpen = new JMenuItem("開く(O)");
		itemOpen.setMnemonic('O');
		itemOpen.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.openFile();
			}
		});
		itemOpen.setIcon(this.getImageIcon("images/open.png"));
		menuFile.add(itemOpen);

		JMenuItem itemSave = new JMenuItem("保存(S)");
		itemSave.setMnemonic('S');
		itemSave.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.saveFile();
			}
		});
		itemSave.setIcon(this.getImageIcon("images/save.png"));
		menuFile.add(itemSave);

		menuFile.addSeparator();

		JMenuItem itemExit = new JMenuItem("終了(X)", 'X');
		itemExit.setIcon(this.getImageIcon("images/close.png"));
		itemExit.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				System.exit(0);
			}
		});
		menuFile.add(itemExit);
		menuBar.add(menuFile);

		// 編集(E)
		JMenu menuEdit = new JMenu("編集(E)");
		menuEdit.setMnemonic('E');

		JMenuItem cut = new JMenuItem("切り取り(T)", 'T');
		cut.setAccelerator(KeyStroke.getKeyStroke('X', InputEvent.CTRL_MASK));
		cut.setIcon(this.getImageIcon("images/cut.png"));
		cut.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.editor.getTextPane().cut();
			}
		});
		menuEdit.add(cut);

		JMenuItem copy = new JMenuItem("コピー(C)", 'C');
		copy.setAccelerator(KeyStroke.getKeyStroke('C', InputEvent.CTRL_MASK));
		copy.setIcon(this.getImageIcon("images/copy.png"));
		copy.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.editor.getTextPane().copy();
			}
		});
		menuEdit.add(copy);

		JMenuItem paste = new JMenuItem("貼り付け(Y)", 'Y');
		paste.setAccelerator(KeyStroke.getKeyStroke('V', InputEvent.CTRL_MASK));
		paste.setIcon(this.getImageIcon("images/paste.png"));
		paste.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.editor.getTextPane().paste();
			}
		});
		menuEdit.add(paste);

		menuBar.add(menuEdit);

		// ビルド(B)
		JMenu menuBuild = new JMenu("ビルド(B)");
		menuBuild.setMnemonic('B');

		JMenuItem assemble = new JMenuItem("アセンブル(A)", 'A');
		assemble.setMnemonic('A');
		assemble.setIcon(this.getImageIcon("images/assemble.png"));
		assemble.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.assemble();
			}
		});
		menuBuild.add(assemble);
		menuBar.add(menuBuild);

		JMenuItem compile = new JMenuItem("コンパイル(B)", 'B');
		compile.setMnemonic('B');
		compile.setIcon(this.getImageIcon("images/compile.png"));
		compile.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.compile();
			}
		});
		menuBuild.add(compile);
		menuBar.add(menuBuild);

		// ヘルプ(H)
		JMenu menuHelp = new JMenu("ヘルプ(H)");
		menuHelp.setMnemonic('H');

		JMenuItem menu = new JMenuItem("ORANGE-4 IDEについて(A)", 'A');
		menu.setIcon(this.getImageIcon("images/about.png"));
		menu.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				ImageIcon icon = Frame.this.getImageIcon("images/ORANGE4.png");
				JOptionPane.showMessageDialog(Frame.this, "Ver 1.01 Copyright(C) 2017 picosoft Co., Ltd.",
						"ORANGE-4 IDE",
						JOptionPane.INFORMATION_MESSAGE, icon);
			}
		});
		menuHelp.add(menu);

		menuBar.add(menuHelp);

		return menuBar;
	}

	/**
	 *
	 * @return
	 */
	private JToolBar createToolBar() {
		JToolBar toolBar = new JToolBar();
		toolBar.setFloatable(false);

		JButton newButton = new JButton(this.getImageIcon("images/new.png"));
		newButton.setToolTipText("新規");
		newButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.newFile();
			}
		});
		toolBar.add(newButton);

		JButton open = new JButton(this.getImageIcon("images/open.png"));
		open.setToolTipText("開く");
		open.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.openFile();
			}
		});
		toolBar.add(open);

		JButton save = new JButton(this.getImageIcon("images/save.png"));
		save.setToolTipText("保存");
		save.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.saveFile();
			}
		});
		toolBar.add(save);

		toolBar.addSeparator();

		JButton cut = new JButton(this.getImageIcon("images/cut.png"));
		cut.setToolTipText("切り取り");
		cut.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.editor.getTextPane().cut();
			}
		});
		toolBar.add(cut);

		JButton copy = new JButton(this.getImageIcon("images/copy.png"));
		copy.setToolTipText("コピー");
		copy.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.editor.getTextPane().copy();
			}
		});
		toolBar.add(copy);

		JButton paste = new JButton(this.getImageIcon("images/paste.png"));
		paste.setToolTipText("貼り付け");
		paste.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.editor.getTextPane().paste();
			}
		});
		toolBar.add(paste);

		toolBar.addSeparator();

		JButton assemble = new JButton(this.getImageIcon("images/assemble.png"));
		assemble.setToolTipText("アセンブル");
		assemble.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.assemble();
			}
		});
		toolBar.add(assemble);

		JButton complie = new JButton(this.getImageIcon("images/compile.png"));
		complie.setToolTipText("コンパイル");
		complie.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.compile();
			}
		});
		toolBar.add(complie);

		toolBar.addSeparator();
		JButton cls = new JButton(this.getImageIcon("images/cls.png"));
		cls.setToolTipText("クリア");
		cls.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				Frame.this.console.cls();
			}
		});
		toolBar.add(cls);

		return toolBar;
	}

	private ImageIcon getImageIcon(String path) {
		Image image = null;
		URL url = this.getClass().getClassLoader().getResource(path);
		if (url == null) {
			return new ImageIcon(path);
		}
		try {
			image = this.createImage((ImageProducer)url.getContent());
		} catch (IOException e) {
			e.printStackTrace();
		}
		return new ImageIcon(image);
	}

	/**
	 *
	 */
	private void newFile() {
		this.editorTab.addTab("新規", this.editor);
		this.editor.setText("");
		this.openFile = null;
	}

	/**
	 *
	 */
	private void openFile() {
		JFileChooser filechooser = new JFileChooser(this.lastSelectedFile);
		int selected = filechooser.showOpenDialog(this);
		if (selected == JFileChooser.APPROVE_OPTION) {
			this.lastSelectedFile = filechooser.getSelectedFile();
			this.openFile = filechooser.getSelectedFile();
			try {
				String path = this.lastSelectedFile.getPath();
				this.editorTab.addTab(path, this.editor);
				String text = this.loadFile(path);
				this.editor.setText(text);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	private void saveFile() {
		File file = this.openFile;
		if (file == null) {
			JFileChooser filechooser = new JFileChooser(this.lastSelectedFile);
			int selected = filechooser.showSaveDialog(this);
			if (selected == JFileChooser.APPROVE_OPTION) {
				file = filechooser.getSelectedFile();
				if (!file.exists()) {
					try {
						file = Files.createFile(file.toPath()).toFile();
					} catch (IOException e) {
						e.printStackTrace();
						return;
					}
				}
			} else {
				return;
			}
		}
		this.editorTab.addTab(file.getPath(), this.editor);
		try {
			Files.write(file.toPath(), this.editor.getText().getBytes(StandardCharsets.UTF_8),
					StandardOpenOption.TRUNCATE_EXISTING);
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void assemble() {
		String fileName = (this.openFile == null) ? "新規" : this.openFile.getPath();
		System.out.println("----------------------------------------");
		System.out.println(fileName);
		System.out.println("----------------------------------------");
		String src = this.editor.getText();
		new Assembler().assemble(src);
		System.out.println("");
	}

	private void compile() {
		// String fileName = (this.openFile == null) ? "新規" :
		// this.openFile.getPath();
	}

	/**
	 *
	 * @param path
	 * @return
	 * @throws IOException
	 */
	private String loadFile(String path) throws IOException {
		return new String(Files.readAllBytes(Paths.get(path)), StandardCharsets.UTF_8);
	}

	/**
	 * ドラッグ＆ドロップ（ドロップ時）イベント用メソッドです。 ドラッグ操作を検出したときに呼び出されます。
	 * 
	 * @param dtde
	 *            ドラッグ＆ドロップイベント
	 */
	@Override
	public void dragEnter(DropTargetDragEvent dtde) {
		dtde.acceptDrag(DnDConstants.ACTION_COPY);
	}

	/**
	 * ドラッグ＆ドロップ（ドロップ時）イベント用メソッドです。 ドラッグ操作がドロップなしのときに呼び出されます。
	 * 
	 * @param dte
	 *            ドラッグ＆ドロップイベント
	 */
	@Override
	public void dragExit(DropTargetEvent dte) {
	}

	/**
	 * ドラッグ＆ドロップ（ドロップ時）イベント用メソッドです。 ドラッグ操作が進行中のときに呼び出されます。
	 * 
	 * @param dte
	 *            ドラッグ＆ドロップイベント
	 */
	@Override
	public void dragOver(DropTargetDragEvent dtde) {
	}

	/**
	 * ドラッグ＆ドロップ（ドロップ時）イベント用メソッドです。 ドラッグ操作がドロップで終了したときに呼び出されます。
	 * 
	 * @param dte
	 *            ドラッグ＆ドロップイベント
	 */
	@Override
	public void drop(DropTargetDropEvent dtde) {
		try {
			Transferable tr = dtde.getTransferable(); // 転送されたデータの取得
			if (tr.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) { // 転送データがサポート可能
				dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE); // ドロップ動作受入れ
				@SuppressWarnings("rawtypes")
				List myList = (List)tr.getTransferData(DataFlavor.javaFileListFlavor);
				if (myList.size() == 1) { // 要素が一つの場合のみ受入れ
					File file = (File)myList.get(0);
					if (!file.isDirectory()) { // ディレクトリーでない
						this.lastSelectedFile = file;
						String path = this.lastSelectedFile.getPath();
						this.editorTab.addTab(path, this.editor);
						String text = this.loadFile(path);
						this.editor.setText(text);
					}
					dtde.getDropTargetContext().dropComplete(true); // ドロップ動作正常終了
				} else { // 要素が一つでない
					dtde.getDropTargetContext().dropComplete(false); // ドロップ動作異常終了
				}
			} else {
				dtde.rejectDrop(); // ドロップ動作をはねつける
			}
		} catch (IOException ioe) {
			dtde.rejectDrop();
		} catch (UnsupportedFlavorException ufe) {
			dtde.rejectDrop();
		}
	}

	/**
	 * ドラッグ＆ドロップ（ドロップ時）イベント用メソッドです。 ユーザが現在のドロップジェスチャーを変更したときに呼び出されます。
	 * 
	 * @param dte
	 *            ドラッグ＆ドロップイベント
	 */
	@Override
	public void dropActionChanged(DropTargetDragEvent dtde) {
	}

}
